@@ -5,12 +5,12 @@
 // except according to those terms.
 
 use std::collections::BTreeSet;
+use std::iter;
 
 use chrono::{DateTime, NaiveDate, Utc};
 use derive_builder::Builder;
-use itertools::Itertools;
 
-use crate::api::common::NameOrId;
+use crate::api::common::{CommaSeparatedList, NameOrId};
 use crate::api::endpoint_prelude::*;
 use crate::api::ParamValue;
 
@@ -33,7 +33,7 @@ impl IssueStateEvent {
 }
 
 impl ParamValue<'static> for IssueStateEvent {
-    fn as_value(self) -> Cow<'static, str> {
+    fn as_value(&self) -> Cow<'static, str> {
         self.as_str().into()
     }
 }
@@ -47,14 +47,14 @@ enum IssueAssignees {
 #[derive(Debug, Clone)]
 enum IssueLabels<'a> {
     Unlabeled,
-    Labeled(BTreeSet<Cow<'a, str>>),
+    Labeled(CommaSeparatedList<Cow<'a, str>>),
 }
 
 impl<'a, 'b: 'a> ParamValue<'a> for &'b IssueLabels<'a> {
-    fn as_value(self) -> Cow<'a, str> {
+    fn as_value(&self) -> Cow<'a, str> {
         match self {
             IssueLabels::Unlabeled => "".into(),
-            IssueLabels::Labeled(labels) => format!("{}", labels.iter().format(",")).into(),
+            IssueLabels::Labeled(labels) => format!("{}", labels).into(),
         }
     }
 }
@@ -85,6 +85,10 @@ pub struct EditIssue<'a> {
     /// Labels to set on the issue.
     #[builder(setter(name = "_labels"), default, private)]
     labels: Option<IssueLabels<'a>>,
+    #[builder(setter(name = "_add_labels"), default, private)]
+    add_labels: Option<CommaSeparatedList<Cow<'a, str>>>,
+    #[builder(setter(name = "_remove_labels"), default, private)]
+    remove_labels: Option<CommaSeparatedList<Cow<'a, str>>>,
     /// Change the state of the issue.
     #[builder(default)]
     state_event: Option<IssueStateEvent>,
@@ -163,7 +167,13 @@ impl<'a> EditIssueBuilder<'a> {
     }
 
     /// Remove all labels from the issue.
+    #[deprecated(note = "use `clear_labels` instead")]
     pub fn remove_labels(&mut self) -> &mut Self {
+        self.clear_labels()
+    }
+
+    /// Remove all labels from the issue.
+    pub fn clear_labels(&mut self) -> &mut Self {
         self.labels = Some(Some(IssueLabels::Unlabeled));
         self
     }
@@ -172,18 +182,18 @@ impl<'a> EditIssueBuilder<'a> {
     ///
     /// Note that the list of labels sent will replace the set on the instance. This only adds it
     /// to the list of labels to add to the set before sending it to the instance.
+    ///
+    /// See: `add_label`.
     pub fn label<L>(&mut self, label: L) -> &mut Self
     where
         L: Into<Cow<'a, str>>,
     {
         let label = label.into();
         let labels = if let Some(Some(IssueLabels::Labeled(mut set))) = self.labels.take() {
-            set.insert(label);
+            set.push(label);
             set
         } else {
-            let mut set = BTreeSet::new();
-            set.insert(label);
-            set
+            iter::once(label).collect()
         };
         self.labels = Some(Some(IssueLabels::Labeled(labels)));
         self
@@ -193,6 +203,8 @@ impl<'a> EditIssueBuilder<'a> {
     ///
     /// Note that the list of labels sent will replace the set on the instance. This only adds it
     /// to the list of labels to add to the set before sending it to the instance.
+    ///
+    /// See: `add_label`.
     pub fn labels<I, L>(&mut self, iter: I) -> &mut Self
     where
         I: IntoIterator<Item = L>,
@@ -208,6 +220,34 @@ impl<'a> EditIssueBuilder<'a> {
         self.labels = Some(Some(IssueLabels::Labeled(labels)));
         self
     }
+
+    /// Add a label to the issue.
+    ///
+    /// This is an incremental addition to the existing set of labels on the issue.
+    pub fn add_label<L>(&mut self, label: L) -> &mut Self
+    where
+        L: Into<Cow<'a, str>>,
+    {
+        self.add_labels
+            .get_or_insert(None)
+            .get_or_insert_with(CommaSeparatedList::new)
+            .push(label.into());
+        self
+    }
+
+    /// Remove a label from the issue.
+    ///
+    /// This is an incremental remove form the existing set of labels on the issue.
+    pub fn remove_label<L>(&mut self, label: L) -> &mut Self
+    where
+        L: Into<Cow<'a, str>>,
+    {
+        self.remove_labels
+            .get_or_insert(None)
+            .get_or_insert_with(CommaSeparatedList::new)
+            .push(label.into());
+        self
+    }
 }
 
 impl<'a> Endpoint for EditIssue<'a> {
@@ -227,6 +267,8 @@ impl<'a> Endpoint for EditIssue<'a> {
             .push_opt("description", self.description.as_ref())
             .push_opt("milestone_id", self.milestone_id)
             .push_opt("labels", self.labels.as_ref())
+            .push_opt("add_labels", self.add_labels.as_ref())
+            .push_opt("remove_labels", self.remove_labels.as_ref())
             .push_opt("state_event", self.state_event)
             .push_opt("updated_at", self.updated_at)
             .push_opt("due_date", self.due_date)
@@ -469,6 +511,7 @@ mod tests {
     }
 
     #[test]
+    #[allow(deprecated)]
     fn endpoint_labels_remove() {
         let endpoint = ExpectedUrl::builder()
             .method(Method::PUT)
@@ -488,6 +531,68 @@ mod tests {
         api::ignore(endpoint).query(&client).unwrap();
     }
 
+    #[test]
+    fn endpoint_labels_clear() {
+        let endpoint = ExpectedUrl::builder()
+            .method(Method::PUT)
+            .endpoint("projects/simple%2Fproject/issues/1")
+            .content_type("application/x-www-form-urlencoded")
+            .body_str("labels=")
+            .build()
+            .unwrap();
+        let client = SingleTestClient::new_raw(endpoint, "");
+
+        let endpoint = EditIssue::builder()
+            .project("simple/project")
+            .issue(1)
+            .clear_labels()
+            .build()
+            .unwrap();
+        api::ignore(endpoint).query(&client).unwrap();
+    }
+
+    #[test]
+    fn endpoint_add_labels() {
+        let endpoint = ExpectedUrl::builder()
+            .method(Method::PUT)
+            .endpoint("projects/simple%2Fproject/issues/1")
+            .content_type("application/x-www-form-urlencoded")
+            .body_str("add_labels=one%2Ctwo")
+            .build()
+            .unwrap();
+        let client = SingleTestClient::new_raw(endpoint, "");
+
+        let endpoint = EditIssue::builder()
+            .project("simple/project")
+            .issue(1)
+            .add_label("one")
+            .add_label("two")
+            .build()
+            .unwrap();
+        api::ignore(endpoint).query(&client).unwrap();
+    }
+
+    #[test]
+    fn endpoint_remove_labels() {
+        let endpoint = ExpectedUrl::builder()
+            .method(Method::PUT)
+            .endpoint("projects/simple%2Fproject/issues/1")
+            .content_type("application/x-www-form-urlencoded")
+            .body_str("remove_labels=one%2Ctwo")
+            .build()
+            .unwrap();
+        let client = SingleTestClient::new_raw(endpoint, "");
+
+        let endpoint = EditIssue::builder()
+            .project("simple/project")
+            .issue(1)
+            .remove_label("one")
+            .remove_label("two")
+            .build()
+            .unwrap();
+        api::ignore(endpoint).query(&client).unwrap();
+    }
+
     #[test]
     fn endpoint_state_event() {
         let endpoint = ExpectedUrl::builder()
